// $Id$
//
// Copyright (C) 2004 Satoshi Imai
//

#include "cpuid.h"

//
// cpuidpbP[W
//

/*
 * VC++ 5.x has no 'cpuid' assembler instruction, so we
 * must emulate it
 */
#if defined(_MSC_VER) && ( _MSC_VER <= 1100 )
#define cpuid __asm __emit 0fh __asm __emit 0a2h
#endif

/*
 *------------------------------------------------------------------------
 *
 * TclWinCPUID --
 *
 *	Get CPU ID information on an Intel box under Windows
 *
 * Results:
 *	Returns TCL_OK if successful, TCL_ERROR if CPUID is not
 *	supported or fails.
 *
 * Side effects:
 *	If successful, stores EAX, EBX, ECX and EDX registers after 
 *      the CPUID instruction in the four integers designated by 'regsPtr'
 *
 *----------------------------------------------------------------------
 */

int
TclWinCPUID( unsigned int index, /* Which CPUID value to retrieve */
	     register unsigned int * regsPtr ) /* Registers after the CPUID */
{

    int status = TCL_ERROR;

#if defined(__GNUC__)

    /* Establish structured exception handling */

# ifdef HAVE_NO_SEH
    __asm__ __volatile__ (
	    "pushl %%ebp" "\n\t"
	    "pushl %0" "\n\t"
	    "pushl %%fs:0" "\n\t"
	    "movl  %%esp, %%fs:0"
            :
            : "r" (_except_TclWinCPUID_detach_handler) );
#  else
    __try {
#  endif

	/* 
	 * Execute the CPUID instruction with the given index, and
	 * store results off 'regPtr'.
	 */

	__asm__ __volatile__ (
	    "movl %4, %%eax" "\n\t"
            "cpuid" "\n\t"
	    "movl %%eax, %0" "\n\t"
	    "movl %%ebx, %1" "\n\t"
	    "movl %%ecx, %2" "\n\t"
	    "movl %%edx, %3"
	    : 
	    "=m"(regsPtr[0]),
	    "=m"(regsPtr[1]),
	    "=m"(regsPtr[2]),
	    "=m"(regsPtr[3])
	    : "m"(index)
	    : "%eax", "%ebx", "%ecx", "%edx" );
	status = TCL_OK;

	/* End the structured exception handler */

#  ifndef HAVE_NO_SEH
    } __except( EXCEPTION_EXECUTE_HANDLER ) {
	/* do nothing */
    }
#  else
    __asm __volatile__ (
	    "jmp  TclWinCPUID_detach_pop" "\n"
        "TclWinCPUID_detach_reentry:" "\n\t"
	    "movl %%fs:0, %%eax" "\n\t"
	    "movl 0x8(%%eax), %%esp" "\n\t"
	    "movl 0x8(%%esp), %%ebp" "\n"
	"TclWinCPUID_detach_pop:" "\n\t"
	    "movl (%%esp), %%eax" "\n\t"
	    "movl %%eax, %%fs:0" "\n\t"
	    "add  $12, %%esp" "\n\t"
	:
	:
	: "%eax");
#  endif


#elif defined(_MSC_VER)

    /* Define a structure in the stack frame to hold the registers */

    struct {
	DWORD dw0;
	DWORD dw1;
	DWORD dw2;
	DWORD dw3;
    } regs;
    regs.dw0 = index;
    
    /* Execute the CPUID instruction and save regs in the stack frame */

    _try {
	_asm {
	    push    ebx
	    push    ecx
	    push    edx
	    mov     eax, regs.dw0
            cpuid
	    mov     regs.dw0, eax
	    mov     regs.dw1, ebx
	    mov     regs.dw2, ecx
	    mov     regs.dw3, edx
            pop     edx
            pop     ecx
            pop     ebx
	}
	
	/* Copy regs back out to the caller */

	regsPtr[0]=regs.dw0;
	regsPtr[1]=regs.dw1;
	regsPtr[2]=regs.dw2;
	regsPtr[3]=regs.dw3;

	status = TCL_OK;
    } __except( EXCEPTION_EXECUTE_HANDLER ) {
    }

#else
				/* Don't know how to do assembly code for
				 * this compiler */
#endif
    return status;
}

/*
 *----------------------------------------------------------------------
 *
 * TestwincpuidCmd --
 *
 *	Retrieves CPU ID information.
 *
 * Usage:
 *	testwincpuid <eax>
 *
 * Parameters:
 *	eax - The value to pass in the EAX register to a CPUID instruction.
 *
 * Results:
 *	Returns a four-element list containing the values from the
 *	EAX, EBX, ECX and EDX registers returned from the CPUID instruction.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
TestwincpuidCmd( ClientData dummy,
		 Tcl_Interp* interp, /* Tcl interpreter */
		 int objc,	/* Parameter count */
		 Tcl_Obj *CONST * objv ) /* Parameter vector */
{
    int status;
    int index;
    unsigned int regs[4];
    Tcl_Obj * regsObjs[4];
    int i;

    if ( objc != 2 ) {
	Tcl_WrongNumArgs( interp, 1, objv, "eax" );
	return TCL_ERROR;
    }
    if ( Tcl_GetIntFromObj( interp, objv[1], &index ) != TCL_OK ) {
	return TCL_ERROR;
    }
    status = TclWinCPUID( (unsigned int) index, regs );
    if ( status != TCL_OK ) {
	Tcl_SetObjResult( interp, Tcl_NewStringObj( "operation not available", 
						    -1 ) );
	return status;
    }
    for ( i = 0; i < 4; ++i ) {
	regsObjs[i] = Tcl_NewIntObj( (int) regs[i] );
    }
    Tcl_SetObjResult( interp, Tcl_NewListObj( 4, regsObjs ) );
    return TCL_OK;
       
}

//
// R}h̒ǉ
//
Tcl_Command
AddCommand(
	Tcl_Interp *interp,
	char *nameSpace,
	char *cmdName,
	Tcl_ObjCmdProc *cmdProc
)
{
	char *cmdPath;
	Tcl_DString dString;
	Tcl_Command cmdToken;

	Tcl_DStringInit(&dString);

	// Namespace::Command 쐬
	if (nameSpace != NULL) {
		Tcl_DStringAppend(&dString, nameSpace, -1);
	}
	Tcl_DStringAppend(&dString, "::", -1);
	Tcl_DStringAppend(&dString, cmdName, -1);

	cmdPath = Tcl_DStringValue(&dString);
	// o^̃`FbN
	cmdToken = Tcl_FindCommand(interp, cmdPath, (Tcl_Namespace *)NULL, 0);
	if (cmdToken != NULL) {
		Tcl_DStringFree(&dString);
		return cmdToken;
	}
	// R}h̓o^
	cmdToken = Tcl_CreateObjCommand(interp, cmdPath, cmdProc, NULL, NULL);
	Tcl_DStringFree(&dString);

	Tcl_Namespace *nsPtr;
	int dontResetList = 0;

	// Namespacẽ|C^擾
	nsPtr = Tcl_FindNamespace(interp, nameSpace, (Tcl_Namespace *)NULL, TCL_LEAVE_ERR_MSG);
	if (nsPtr == NULL) {
		return NULL;
	}
	// R}hGNX|[g
	if (Tcl_Export(interp, nsPtr, cmdName, dontResetList) != TCL_OK) {
		return NULL;
	}
	return cmdToken;
}

//
// gDLL̏
//
int
Cpuid_Init(
	Tcl_Interp *interp
)
{
#ifdef USE_TCL_STUBS
	// Tcl Stub̏
	if (Tcl_InitStubs(interp, "8.1", 0) == NULL) {
		return TCL_ERROR;
	}
#endif // USE_TCL_STUBS
#ifdef USE_TK_STUBS
	// Tk Stub̏
	if (Tk_InitStubs(interp, "8.1", 0) == NULL) {
		return TCL_ERROR;
	}
#endif // USE_TK_STUBS

	// Namespace̐
	Tcl_Namespace *spacePtr = Tcl_CreateNamespace(interp, NAMESPACE, (ClientData)0, (Tcl_NamespaceDeleteProc *) NULL);
	if (spacePtr == NULL) {
		return TCL_ERROR;
	}

	// R}h̒ǉ
	AddCommand(interp, NAMESPACE, "cpuid", TestwincpuidCmd);

	// pbP[W̓o^ƃo[W̕ԋp
	return Tcl_PkgProvide(interp, PACKAGE, VERSION);
}
